@@ -10,16 +10,12 @@ import attrdict
 from getpass import getpass
 from queue import Queue
 from threading import *
+from colorama import Fore, init, Style
+
+init()
 
 class Bot:
 
-    @staticmethod
-    def log(message, error = False, _copy = False):
-        ERROR = "\033[1;31;40m"
-        NORMAL = "\033[m"
-        print(f'{ERROR if error else NORMAL}{"COPY: " if _copy else ""}{message}{NORMAL}')
-        sys.stdout.flush()
-
     @staticmethod
     def randomize_between(*responses):
         response = random.choice(responses)
@@ -39,7 +35,8 @@ class Bot:
             "APITimeout",
             "advanced",
             "owner",
-            "important"]
+            "important",
+            "killChildren"]
         noneVars = ["userEmail",
             "userPassword",
             "help",
@@ -63,6 +60,7 @@ class Bot:
         self.human = False
         self._regexes = {}
         self._copy = False
+        self.killChildren = False
         for item in vars:
             if item in kwargs:
                 setattr(self, item, kwargs[item])
@@ -81,32 +79,40 @@ class Bot:
             self.room = self.args.room if self.args.test != True or self._copy else "test"
             self.password = self.args.password
             if not self._copy:
-                Bot.log(f"Debug: {str(self.args.test)}", _copy = self._copy)
+                self.log(f"Debug: {str(self.args.test)}")
 
     def _wait_for(self, response):
-        reply = None
-        i = 0
-        waitlist = Queue()
-        for msgJSON in self.conn:
-            i += 1
-            msg = AttrDict(json.loads(msgJSON))
-            if msg.type == response:
-                reply = msg
-                break
-            else:
-                waitlist.put(msg)
-            if i > self.APITimeout:
-                Bot.log(f"{response} API response not recorded. reason: too many events before {response}.", error = True, _copy = self._copy)
-                reply = None
-                break
-        while not waitlist.empty():
-            self._handle(waitlist.get())
-        return reply
+        if self.conn.sock:
+            reply = None
+            i = 0
+            waitlist = Queue()
+            for msgJSON in self.conn:
+                i += 1
+                msg = AttrDict(json.loads(msgJSON))
+                if msg.type == response:
+                    reply = msg
+                    break
+                else:
+                    waitlist.put(msg)
+                if i > self.APITimeout:
+                    self.log(f"{response} API response not recorded. reason: too many events before {response}.", error = True)
+                    reply = None
+                    break
+            while not waitlist.empty():
+                self._handle(waitlist.get())
+            return reply
+        else:
+            return None
 
     def _recv(self):
-        return AttrDict(json.loads(self.conn.recv()))
+        try:
+            return AttrDict(json.loads(self.conn.recv()))
+        except json.decoder.JSONDecodeError:
+            self.kill()
 
     def _handle(self, msg):
+        if msg is None:
+            return
         if msg.type == 'send-event':
             if re.search(f'^!kill @{re.escape(self.normname)}$', msg.data.content) is not None:
                 self._handle_kill(msg)
@@ -114,16 +120,16 @@ class Bot:
                 self.kill()
             if re.search(f'^!restart @{re.escape(self.normname)}$', msg.data.content) is not None:
                 self.restart(msg = msg)
-                raise KilledError(f"old {self.nick} in &{self.room} killed. Restarted.")
+                raise KilledError(bot = self, message = f"old {self.nick} in &{self.room} killed. Restarted.")
         if msg.type == 'error':
-            Bot.log(msg, error = True, _copy = self._copy)
+            self.log(msg, error = True)
         elif msg.type in self.handlers:
             self.handlers[msg.type](msg)
         else:
             if callable(self.function):
                 self.function(msg)
             else:
-                Bot.log("Advanced start must be given a callable message handler function that takes an AttrDict as its argument.", error = True, _copy = self._copy)
+                self.log("Advanced start must be given a callable message handler function that takes an AttrDict as its argument.", error = True)
 
     def _is_privileged(self, user):
         if "is_manager" in user or user.name == self.owner:
@@ -157,7 +163,7 @@ class Bot:
                             self.send_msg(send, parent = msg)
                 else:
                     if isinstance(response, str):
-                        self.send_msg(msgString = result, parent = msg)
+                        self.send_msg(msgString = response, parent = msg)
                     elif isinstance(response, dict):
                         for send, nick in response.items():
                             self.set_nick(nick)
@@ -167,17 +173,12 @@ class Bot:
                         for send in response:
                             self.send_msg(send, parent = msg)
                     else:
-                        raise BadMessageError(f"Response type: {type(response)}")
+                        raise BadMessageError(message = f"Response type: {type(response)}", bot = self)
                 break
 
     def _handle_kill(self, msg):
         if self._is_privileged(msg.data.sender) or not self.important:
             self.send_msg(self.killed, msg)
-            for process in self.childThreads:
-                process.terminate()
-            for copy in self.children:
-                copy.send_msg("/me was killed via parricide")
-                copy.kill()
             self.kill()
         else:
             self.send_msg(f"Bot not killed because you are not a host or @{self.normowner}, and this bot is marked as important.\nFor emergencies, use !forcekill.", parent = msg)
@@ -187,11 +188,15 @@ class Bot:
         reply = self._wait_for("auth-reply")
         if reply is not None:
             if reply.data.success == True:
-                Bot.log(f'Successfully logged into {self.room}.', _copy = self._copy)
+                self.log(f'Successfully logged into {self.room}.')
             else:
-                Bot.log(f'Login unsuccessful. Reason: {reply.data.reason}', error = True, _copy = self._copy)
+                self.log(f'Login unsuccessful. Reason: {reply.data.reason}', error = True)
                 self._handle_auth(getpass("Enter the correct password: "))
 
+    def log(self, message, error = False):
+        print(f'{Fore.RED if error else Style.RESET_ALL}{"COPY: " if self._copy else ""}{message}{Style.RESET_ALL}')
+        sys.stdout.flush()
+
     def connect(self):
         self.conn = websocket.create_connection(f'wss://euphoria.io/room/{self.room}/ws{"?h=0" if not self.human else "?h=1"}', cookie = self._cookie, enable_multithread = True)
         if self._cookie is None:
@@ -206,42 +211,10 @@ class Bot:
             self._handle_auth(self.password)
             self.set_nick(self.nick)
         else:
-            Bot.log(reply, error = True, _copy = self._copy)
-        Bot.log(f'connected to &{self.room}.', _copy = self._copy)
+            self.log(reply, error = True)
+        self.log(f'connected to &{self.room}.')
         return hello
 
-    def send_msg(self, msgString, parent = None):
-        try:
-            if isinstance(parent, AttrDict):
-                self.send_command(data = {'content': msgString, 'parent': parent.data.id})
-            elif isinstance(parent, str):
-                self.send_command(data = {'content': msgString, 'parent': parent})
-            elif parent is None:
-                self.send_command(data = {'content': msgString})
-            else:
-                raise BadMessageError(f'message not sent, because type of parent was {type(parent)}. \nParent printed: \n{parent}')
-            reply = self._wait_for("send-reply")
-            if isinstance(parent, AttrDict):
-                Bot.log(f'Message sent: {reply} replying to: {parent.data.id} by {parent.data.sender.name}', _copy = self._copy)
-            elif isinstance(parent, str):
-                Bot.log(f'Message sent: {reply} replying to: {parent}', _copy = self._copy)
-            elif parent is None:
-                Bot.log(f'Message sent: {reply}', _copy = self._copy)
-        except BadMessageError:
-            pass
-
-    def restart(self, msg = None, advanced = False):
-        if self.conn is not None:
-            self.conn.close(timeout = 0)
-            del self.conn
-            Bot.log("restarting...", _copy = self._copy)
-        else:
-            Bot.log("starting...", _copy = self._copy)
-        self.connect()
-        t = Thread(target = self.start, kwargs = {"advanced" : advanced})
-        t.start()
-        t.join(0)
-
     def start(self, function = None, advanced = False):
         if not advanced and not self.advanced:
             self.set_handler('ping-event', self._handle_ping)
@@ -252,15 +225,57 @@ class Bot:
             if self.function is None and function is not None and callable(function):
                 self.function = function
             else:
-                Bot.log("No valid function passed to advanced bot's start function. Using filler function", _copy = self._copy)
+                self.log("No valid function passed to advanced bot's start function. Using filler function")
                 self.function = lambda msg: None
+            try:
+                while True:
+                    msg = self._recv()
+                    self._handle(msg)
+            except KilledError:
+                pass
+
+    def kill(self):
+        if self.killChildren:
+            for child in self.children:
+                if copy.conn is not None and copy.conn.sock:
+                    copy.conn.close()
+                    copy.conn = None
+        if self.conn is not None and self.conn.sock:
+            self.conn.close()
+        self.conn = None
+        raise KilledError(message = f'{self.nick} in &{self.room} killed.', bot = self)
+
+    def send_msg(self, msgString, parent = None):
         try:
-            for msgJSON in self.conn:
-                msg = AttrDict(json.loads(msgJSON))
-                self._handle(msg)
-        except KilledError:
+            if isinstance(parent, AttrDict):
+                self.send_command(data = {'content': msgString, 'parent': parent.data.id})
+                reply = self._wait_for("send-reply")
+                self.log(f'Message sent: {reply} replying to: {parent.data.id} by {parent.data.sender.name}')
+            elif isinstance(parent, str):
+                self.send_command(data = {'content': msgString, 'parent': parent})
+                reply = self._wait_for("send-reply")
+                self.log(f'Message sent: {reply} replying to: {parent}')
+            elif parent is None:
+                self.send_command(data = {'content': msgString})
+                reply = self._wait_for("send-reply")
+                self.log(f'Message sent: {reply}')
+            else:
+                raise BadMessageError(message = f'message not sent, because type of parent was {type(parent)}. \nParent printed: \n{parent}', bot = self)
+        except BadMessageError:
             pass
 
+    def restart(self, msg = None, advanced = False):
+        if self.conn is not None:
+            self.conn.close(timeout = 0)
+            self.conn = None
+            self.log("restarting...")
+        else:
+            self.log("starting...")
+        self.connect()
+        t = Thread(target = self.start, kwargs = {"advanced" : advanced}, name = f"{self.nick} in &{self.room}")
+        t.start()
+        t.join(0)
+
     def add_command_line_arg(self, *args, **kwargs):
         self.parser.add_argument(*args, **kwargs)
         self.args = self.parser.parse_args()
@@ -295,12 +310,6 @@ class Bot:
     def set_nick(self, nick):
         self.send_command(type = 'nick', data = {'name': nick})
 
-    def kill(self):
-        if self.conn is not None and self.conn.connected:
-            self.conn.close()
-        del self.conn
-        raise KilledError(f'{self.normname} in &{self.room} killed.')
-
     def get_userlist(self):
         self.send_command(type = 'who')
         reply = self._wait_for("who-reply")
@@ -312,13 +321,13 @@ class Bot:
         self.password = password
         self.restart()
         if _copyStart:
-            Bot.log(f"Copy started in &{self.room}.", _copy = self._copy)
+            self.log(f"Copy started in &{self.room}.")
 
     def set_handler(self, eventString, function):
         if callable(function):
             self.handlers[eventString] = function
         else:
-            Bot.log(f"WARNING: handler for {eventString} not callable, handler not set.", error = True, _copy = self._copy)
+            self.log(f"WARNING: handler for {eventString} not callable, handler not set.", error = True)
 
     def set_handlers(self, handlers):
         for eventString, function in handlers.items():
@@ -344,7 +353,7 @@ class Bot:
             human = self.human)
         copy.set_handlers(self.handlers)
         copy.set_regexes(self._regexes)
-        t = Thread(target = copy.move_to, kwargs = {"roomName": roomName, "password" : password, "_copyStart" : True})
+        t = Thread(target = copy.move_to, kwargs = {"roomName": roomName, "password" : password, "_copyStart" : True}, name = f"{copy.nick} in &{copy.room}")
         self.children.append(copy)
         copy.parents.append(self)
         self.childThreads.append(t)
@@ -371,14 +380,14 @@ class Bot:
             if msg.type == "login-reply":
                 reply = msg
                 if reply.data.success:
-                    Bot.log("Login successful.", _copy = self._copy)
+                    self.log("Login successful.")
                     self.account = reply.data.account_id
                 else:
-                    Bot.log(f"Access denied. Reason: {reply.data.reason}", error = True, _copy = self._copy)
+                    self.log(f"Access denied. Reason: {reply.data.reason}", error = True)
                     self.login(email = input("Enter your email: "), password = getpass("Enter your password: "), setDefaults = setDefaults)
                 return reply
             if i > self.APITimeout:
-                Bot.log("login-reply API response not recorded. reason: too many events before pm-initiate-reply.", error = True, _copy = self._copy)
+                self.log("login-reply API response not recorded. reason: too many events before pm-initiate-reply.", error = True)
                 return None
 
     def initiate_pm(self, id, bot = None, room = "pmtest", hostEmail = None, hostPassword = None):
@@ -386,7 +395,7 @@ class Bot:
         copy.connect()
         copy.login(email = hostEmail, password = hostPassword, setDefaults = False)
         copy.conn.close()
-        del copy.conn
+        copy.conn = None
         copy.connect()
         self.send_command(type = 'pm-initiate', data = {'user_id' : id})
         to = self._wait_for("pm-initiate-reply")
@@ -397,18 +406,18 @@ class Bot:
                 else:
                     self.copy_to("pm:{to.data.pm_id}")
             else:
-                Bot.log("Access Denied: This bot does not have permission to initiate PMs.", error = True, _copy = self._copy)
+                self.log("Access Denied: This bot does not have permission to initiate PMs.", error = True)
         return to
 
 class BotError(Exception):
     pass
 
 class KilledError(BotError):
-    def __init__(self, message = "bot killed."):
+    def __init__(self, bot, message = "bot killed."):
         self.message = message
-        Bot.log(f'KilledError: {message}', error = True)
+        bot.log(f'KilledError: {message}', error = True)
 
 class BadMessageError(BotError):
-    def __init__(self, message):
+    def __init__(self, bot, message):
         self.message = message
-        Bot.log(f'BadMessageError: {message}', error = True)
+        bot.log(f'BadMessageError: {message}', error = True)
